from .i18n import _
from .error import OperationError, DictConsistencyError
from .config import Config


class Path:
    """Helper class to handle the `path` attribute of a CreoleObjSpace
    instance.

    sample: path="creole.general.condition"
    """
    def __init__(self):
        self.variables = {}
        self.families = {}
        self.full_paths_families = {}
        self.full_paths_variables = {}

    # Family
    def add_family(self,
                   namespace: str,
                   name: str,
                   variableobj: str,
                   ) -> str:  # pylint: disable=C0111
        if '.' not in name and namespace == Config['variable_namespace']:
            full_name = '.'.join([namespace, name])
            self.full_paths_families[name] = full_name
        else:
            full_name = name
        if full_name in self.families and self.families[full_name]['variableobj'] != variableobj:  # pragma: no cover
            raise DictConsistencyError(_(f'Duplicate family name "{name}"'), 37)
        self.families[full_name] = dict(name=name,
                                        namespace=namespace,
                                        variableobj=variableobj,
                                        )

    def _get_family(self,
                    name: str,
                    namespace: str=None,
                    ):
        # if main namespace, get full_path
        if '.' not in name and namespace in [None, Config['variable_namespace']] and name in self.full_paths_families:
            name = self.full_paths_families[name]
        dico = self.families[name]
        if namespace and dico['namespace'] != Config['variable_namespace'] and namespace != dico['namespace']:
            raise DictConsistencyError(_(f'A family located in the "{dico["namespace"]}" namespace shall not be used in the "{namespace}" namespace'), 38)
        return dico

    def get_family_path(self,
                        name: str,
                        namespace: str,
                        ) -> str:  # pylint: disable=C0111
        return self._get_family(name,
                                namespace,
                                )['name']

    def get_family_obj(self,
                       name: str,
                       ) -> 'Family':  # pylint: disable=C0111
        return self._get_family(name)['variableobj']

    def family_is_defined(self,
                          name: str,
                          ) -> str:  # pylint: disable=C0111
        try:
            self._get_family(name)
            return True
        except KeyError:
            return False

    # Leadership
    def set_leader(self,
                   namespace: str,
                   leader_family_name: str,
                   name: str,
                   leader_name: str,
                   ) -> None:  # pylint: disable=C0111
        # need rebuild path and move object in new path
        old_path = namespace + '.' + leader_family_name + '.' + name
        new_path = namespace + '.' + leader_family_name + '.' + leader_name + '.' + name
        self.variables[new_path] = self.variables.pop(old_path)
        self.variables[new_path]['leader'] = leader_name
        self.variables[new_path]['variableobj'].path = new_path
        if namespace == Config['variable_namespace']:
            self.full_paths_variables[name] = new_path

    def get_leader(self, name):  # pylint: disable=C0111
        return self._get_variable(name)['leader']

    # Variable
    def add_variable(self,
                     namespace: str,
                     name: str,
                     family: str,
                     is_dynamic: bool,
                     variableobj,
                     ) -> str:  # pylint: disable=C0111
        if '.' not in name:
            full_path = '.'.join([namespace, family, name])
            if namespace == Config['variable_namespace']:
                self.full_paths_variables[name] = full_path
        else:
            full_path = name
        variableobj.path = full_path
        self.variables[full_path] = dict(name=name,
                                         family=family,
                                         namespace=namespace,
                                         leader=None,
                                         is_dynamic=is_dynamic,
                                         variableobj=variableobj,
                                         )

    def get_variable_name(self,
                          name: str,
                          ):  # pylint: disable=C0111
        return self._get_variable(name)['name']

    def get_variable_obj(self,
                         name: str,
                         ) -> 'Variable':  # pylint: disable=C0111
        return self._get_variable(name)['variableobj']

    def get_variable_family_name(self,
                                 name: str,
                                 ) -> str:  # pylint: disable=C0111
        return self._get_variable(name)['family']

    def get_variable_namespace(self,
                               name: str,
                               ) -> str:  # pylint: disable=C0111
        return self._get_variable(name)['namespace']

    def get_variable_path(self,
                          name: str,
                          current_namespace: str,
                          allow_source: str=False,
                          with_suffix: bool=False,
                          ) -> str:  # pylint: disable=C0111
        if current_namespace is None:  # pragma: no cover
            raise OperationError('current_namespace must not be None')
        if with_suffix:
            dico, suffix = self._get_variable(name,
                                              with_suffix=True,
                                              )
        else:
            dico = self._get_variable(name)
        if not allow_source and dico['namespace'] not in [Config['variable_namespace'], 'services'] and current_namespace != dico['namespace']:
            raise DictConsistencyError(_(f'A variable located in the "{dico["namespace"]}" namespace shall not be used in the "{current_namespace}" namespace'), 41)
        list_path = [dico['namespace'], dico['family']]
        if dico['leader'] is not None:
            list_path.append(dico['leader'])
        list_path.append(dico['name'])
        value = '.'.join(list_path)
        if with_suffix:
            return value, suffix
        return value

    def path_is_defined(self,
                        name: str,
                        ) -> str:  # pylint: disable=C0111
        if '.' not in name and name not in self.variables and name in self.full_paths_variables:
            return True
        return name in self.variables

    def variable_is_dynamic(self,
                            name: str,
                            ) -> bool:
        return self._get_variable(name)['is_dynamic']

    def _get_variable(self,
                      name: str,
                      with_suffix: bool=False,
                      ) -> str:
        if name not in self.variables:
            if '.' not in name and name in self.full_paths_variables:
                name = self.full_paths_variables[name]
            elif with_suffix:
                for var_name, full_path in self.full_paths_variables.items():
                    if name.startswith(var_name):
                        variable = self._get_variable(full_path)
                        if variable['is_dynamic']:
                            return variable, name[len(var_name):]
        if name not in self.variables:
            raise DictConsistencyError(_('unknown option {}').format(name), 42)
        if with_suffix:
            return self.variables[name], None
        return self.variables[name]
